@@ -4,19 +4,30 @@ use strict;
 use warnings;
 
 # ABSTRACT: validator for BibTeX format
-our $VERSION = '0.1.1'; # VERSION
+our $VERSION = '0.2.0'; # VERSION
 
 use Algorithm::CheckDigits;
 use Data::Validate::Email qw( is_email_rfc822 );
 use Data::Validate::URI qw( is_uri );
 use Scalar::Util qw( blessed );
+use Text::BibTeX::Validate::Warning;
 
 require Exporter;
 our @ISA = qw( Exporter );
 our @EXPORT_OK = qw(
+    shorten_DOI
     validate_BibTeX
 );
 
+sub shorten_DOI($)
+{
+    my( $doi ) = @_;
+
+    return $doi if $doi =~ s|^https?://(dx\.)?doi\.org/||;
+    return $doi if $doi =~ s|^doi:||;
+    return $doi;
+}
+
 sub validate_BibTeX
 {
     my( $what ) = @_;
@@ -28,30 +39,43 @@ sub validate_BibTeX
     # TODO: check for duplicated keys
     my $entry = { map { lc $_ => $what->{$_} } keys %$what };
 
+    # Report and remove empty keys
+    for my $key (sort keys %$entry) {
+        next if defined $entry->{$key};
+        _warn_value( 'undefined value', $entry, $key );
+        delete $entry->{$key};
+    }
+
     if( exists $entry->{email} &&
         !defined is_email_rfc822 $entry->{email} ) {
-        warn sprintf 'email: value \'%s\' does not look like valid ' .
-                     'email address' . "\n",
-                     $entry->{email};
+        _warn_value( 'value \'%(value)s\' does not look like valid ' .
+                     'email address',
+                     $entry,
+                     'email' );
     }
 
     if( exists $entry->{doi} ) {
         my $doi = $entry->{doi};
-        if( $entry->{doi} =~ m|^https?://doi\.org/(10\.[^/]+/.*)$| ) {
-            warn sprintf 'doi: value \'%s\' is better written as \'%s\'' . "\n",
-                         $entry->{doi},
-                         $1;
-        } elsif( $entry->{doi} !~ m|^(doi:)?10\.[^/]+/.*| ) {
-            warn sprintf 'doi: value \'%s\' does not look like valid DOI' . "\n",
-                         $entry->{doi};
+        my $doi_now = shorten_DOI $doi;
+
+        if( $doi_now !~ m|^10\.[^/]+/| ) {
+            _warn_value( 'value \'%(value)s\' does not look like valid DOI',
+                         $entry,
+                         'doi' );
+        } elsif( $doi ne $doi_now ) {
+            _warn_value( 'value \'%(value)s\' is better written as \'%(suggestion)s\'',
+                         $entry,
+                         'doi',
+                         { suggestion => $doi_now } );
         }
     }
 
     # Validated according to BibTeX recommendations
     if( exists $entry->{month} &&
         $entry->{month} !~ /^(jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)\.?$/i ) {
-        warn sprintf 'month: value \'%s\' does not look like valid month' . "\n",
-                     $entry->{month};
+        _warn_value( 'value \'%(value)s\' does not look like valid month',
+                     $entry,
+                     'month' );
     }
 
     if( exists $entry->{year} ) {
@@ -59,11 +83,13 @@ sub validate_BibTeX
         # are going to be published soon.
         my @localtime = localtime;
         if( $entry->{year} !~ /^[0-9]{4}$/ ) {
-            warn sprintf 'year: value \'%s\' does not look like valid year' . "\n",
-                         $entry->{year};
+            _warn_value( 'value \'%(value)s\' does not look like valid year',
+                         $entry,
+                         'year' );
         } elsif( $entry->{year} > $localtime[5] + 1901 ) {
-            warn sprintf 'year: value \'%s\' is too far in the future' . "\n",
-                         $entry->{year};
+            _warn_value( 'value \'%(value)s\' is too far in the future',
+                         $entry,
+                         'year' );
         }
     }
 
@@ -79,30 +105,56 @@ sub validate_BibTeX
             }
         }
         next if $check->is_valid( $entry->{$key} );
-        warn sprintf '%s: value \'%s\' does not look like valid %s' . "\n",
+        _warn_value( 'value \'%(value)s\' does not look like valid %(FIELD)s',
+                     $entry,
                      $key,
-                     $entry->{$key},
-                     uc $key;
+                     { FIELD => uc $key } );
     }
 
     # Both keys are nonstandard
     for my $key ('eprint', 'url') {
         next if !exists $entry->{$key};
         next if defined is_uri $entry->{$key};
-        warn sprintf '%s: value \'%s\' does not look like valid URL' . "\n",
-                     $key,
-                     $entry->{$key};
+
+        if( $entry->{$key} =~ /^(.*)\n$/ && defined is_uri $1 ) {
+            # BibTeX converted from YAML (i.e., Debian::DEP12) might
+            # have trailing newline character attached.
+            _warn_value( 'URL has trailing newline character',
+                         $entry,
+                         $key );
+            next;
+        }
+
+        _warn_value( 'value \'%(value)s\' does not look like valid URL',
+                     $entry,
+                     $key );
     }
 
     # Nonstandard
     if( exists $entry->{pmid} ) {
-        if( $entry->{pmid} !~ /^[1-9][0-9]*$/ ) {
-            warn sprintf 'pmid: value \'%s\' does not look like valid PMID' . "\n",
-                         $entry->{pmid};
+        if( $entry->{pmid} =~ /^PMC[0-9]{7}$/ ) {
+            _warn_value( 'PMCID \'%(value)s\' is provided instead of PMID',
+                         $entry,
+                         'pmid' );
+        } elsif( $entry->{pmid} !~ /^[1-9][0-9]*$/ ) {
+            _warn_value( 'value \'%(value)s\' does not look like valid PMID',
+                         $entry,
+                         'pmid' );
         }
     }
 }
 
+sub _warn_value
+{
+    my( $message, $entry, $field, $extra ) = @_;
+    $extra = {} unless $extra;
+    warn Text::BibTeX::Validate::Warning->new(
+            $message,
+            { field => $field,
+              value => $entry->{$field},
+              %$extra } );
+}
+
 1;
 
 __END__
